@@ -47,6 +47,7 @@ class Mechanize::HTTP::Agent
   attr_reader :digest_challenges # :nodoc:
   attr_accessor :user
   attr_accessor :password
+  attr_accessor :domain
 
   # :section: Redirection
 
@@ -77,37 +78,11 @@ class Mechanize::HTTP::Agent
 
   # :section: SSL
 
-  # Path to an OpenSSL server certificate file
-  attr_accessor :ca_file
-
-  # An OpenSSL private key or the path to a private key
-  attr_accessor :key
-
-  # An OpenSSL client certificate or the path to a certificate file.
-  attr_accessor :cert
-
-  # An SSL certificate store
-  attr_accessor :cert_store
-
   # OpenSSL key password
   attr_accessor :pass
 
-  # A callback for additional certificate verification.  See
-  # OpenSSL::SSL::SSLContext#verify_callback
-  #
-  # The callback can be used for debugging or to ignore errors by always
-  # returning +true+.  Specifying nil uses the default method that was valid
-  # when the SSLContext was created
-  attr_accessor :verify_callback
-
-  # How to verify SSL connections.  Defaults to VERIFY_PEER
-  attr_accessor :verify_mode
-
   # :section: Timeouts
 
-  # Reset connections that have not been used in this many seconds
-  attr_reader   :idle_timeout
-
   # Set to false to disable HTTP/1.1 keep-alive requests
   attr_accessor :keep_alive
 
@@ -122,14 +97,8 @@ class Mechanize::HTTP::Agent
   # The cookies for this agent
   attr_accessor :cookie_jar
 
-  # URI for a proxy connection
-  attr_reader :proxy_uri
-
-  # Retry non-idempotent requests?
-  attr_reader :retry_change_requests
-
   # Responses larger than this will be written to a Tempfile instead of stored
-  # in memory.
+  # in memory.  Setting this to nil disables creation of Tempfiles.
   attr_accessor :max_file_buffer
 
   # :section: Utility
@@ -156,19 +125,15 @@ class Mechanize::HTTP::Agent
     @follow_meta_refresh_self = false
     @gzip_enabled             = true
     @history                  = Mechanize::History.new
-    @idle_timeout             = nil
     @keep_alive               = true
-    @keep_alive_time          = 300
-    @max_file_buffer          = 10240
+    @max_file_buffer          = 100_000 # 5MB for response bodies
     @open_timeout             = nil
     @post_connect_hooks       = []
     @pre_connect_hooks        = []
-    @proxy_uri                = nil
     @read_timeout             = nil
     @redirect_ok              = true
     @redirection_limit        = 20
     @request_headers          = {}
-    @retry_change_requests    = false
     @robots                   = false
     @user_agent               = nil
     @webrobots                = nil
@@ -184,15 +149,10 @@ class Mechanize::HTTP::Agent
     @digest_challenges    = {}
     @password             = nil # HTTP auth password
     @user                 = nil # HTTP auth user
+    @domain               = nil # NTLM HTTP domain
 
     # SSL
-    @ca_file         = nil
-    @cert            = nil
-    @cert_store      = nil
-    @key             = nil
-    @pass            = nil
-    @verify_callback = nil
-    @verify_mode     = nil
+    @pass = nil
 
     @scheme_handlers = Hash.new { |h, scheme|
       h[scheme] = lambda { |link, page|
@@ -204,6 +164,10 @@ class Mechanize::HTTP::Agent
     @scheme_handlers['https']     = @scheme_handlers['http']
     @scheme_handlers['relative']  = @scheme_handlers['http']
     @scheme_handlers['file']      = @scheme_handlers['http']
+
+    @http = Net::HTTP::Persistent.new 'mechanize'
+    @http.idle_timeout = 5
+    @http.keep_alive   = 300
   end
 
   # Retrieves +uri+ and parses it into a page or other object according to
@@ -264,14 +228,15 @@ class Mechanize::HTTP::Agent
     response = connection.request(uri, request) { |res|
       response_log res
 
-      response_body_io = response_read res, request
+      response_body_io = response_read res, request, uri
 
       res
     }
 
     hook_content_encoding response, uri, response_body_io
 
-    response_body_io = response_content_encoding response, response_body_io
+    response_body_io = response_content_encoding response, response_body_io if
+      request.response_body_permitted?
 
     post_connect uri, response, response_body_io
 
@@ -304,11 +269,21 @@ class Mechanize::HTTP::Agent
     end
   end
 
+  # URI for a proxy connection
+
+  def proxy_uri
+    @http.proxy_uri
+  end
+
+  # Retry non-idempotent requests?
+  def retry_change_requests
+    @http.retry_change_requests
+  end
+
   # Retry non-idempotent requests
 
   def retry_change_requests= retri
-    @retry_change_requests = retri
-    @http.retry_change_requests = retri if @http
+    @http.retry_change_requests = retri
   end
 
   # :section: Headers
@@ -392,6 +367,58 @@ class Mechanize::HTTP::Agent
     end
   end
 
+  ##
+  # Decodes a gzip-encoded +body_io+.  If it cannot be decoded, inflate is
+  # tried followed by raising an error.
+
+  def content_encoding_gunzip body_io
+    log.debug('gzip response') if log
+
+    zio = Zlib::GzipReader.new body_io
+    out_io = auto_io 'mechanize-gunzip', 16384, zio
+    zio.finish
+
+    return out_io
+  rescue Zlib::Error
+    log.error('unable to gunzip response, trying raw inflate') if log
+
+    body_io.rewind
+    body_io.read 10
+
+    begin
+      return inflate body_io, -Zlib::MAX_WBITS
+    rescue Zlib::Error => e
+      log.error("unable to gunzip response: #{e}") if log
+      raise
+    end
+  ensure
+    zio.close if zio and not zio.closed?
+    body_io.close unless body_io.closed?
+  end
+
+  ##
+  # Decodes a deflate-encoded +body_io+.  If it cannot be decoded, raw inflate
+  # is tried followed by raising an error.
+
+  def content_encoding_inflate body_io
+    log.debug('deflate body') if log
+
+    return inflate body_io
+  rescue Zlib::Error
+    log.error('unable to inflate response, trying raw deflate') if log
+
+    body_io.rewind
+
+    begin
+      return inflate body_io, -Zlib::MAX_WBITS
+    rescue Zlib::Error => e
+      log.error("unable to inflate response: #{e}") if log
+      raise
+    end
+  ensure
+    body_io.close
+  end
+
   def disable_keep_alive request
     request['connection'] = 'close' unless @keep_alive
   end
@@ -491,11 +518,17 @@ class Mechanize::HTTP::Agent
     end
   end
 
+  # Sets a Referer header.  Fragment part is removed as demanded by
+  # RFC 2616 14.36, and user information part is removed just like
+  # major browsers do.
   def request_referer request, uri, referer
     return unless referer
     return if 'https' == referer.scheme.downcase and
               'https' != uri.scheme.downcase
-
+    if referer.fragment || referer.user || referer.password
+      referer = referer.dup
+      referer.fragment = referer.user = referer.password = nil
+    end
     request['Referer'] = referer
   end
 
@@ -504,10 +537,24 @@ class Mechanize::HTTP::Agent
   end
 
   def resolve(uri, referer = current_page)
-    uri = uri.dup if uri.is_a?(URI)
+    referer_uri = referer && referer.uri
+    if uri.is_a?(URI)
+      uri = uri.dup
+    elsif uri.nil?
+      if referer_uri
+        return referer_uri
+      end
+      raise ArgumentError, "absolute URL needed (not nil)"
+    else
+      url = uri.to_s.strip
+      if url.empty?
+        if referer_uri
+          return referer_uri.dup.tap { |u| u.fragment = nil }
+        end
+        raise ArgumentError, "absolute URL needed (not #{uri.inspect})"
+      end
 
-    unless uri.is_a?(URI)
-      uri = uri.to_s.strip.gsub(/[^#{0.chr}-#{126.chr}]/o) { |match|
+      url.gsub!(/[^#{0.chr}-#{126.chr}]/o) { |match|
         if RUBY_VERSION >= "1.9.0"
           Mechanize::Util.uri_escape(match)
         else
@@ -515,28 +562,25 @@ class Mechanize::HTTP::Agent
         end
       }
 
-      unescaped = uri.split(/(?:%[0-9A-Fa-f]{2})+|#/)
-      escaped   = uri.scan(/(?:%[0-9A-Fa-f]{2})+|#/)
-
-      escaped_uri = Mechanize::Util.html_unescape(
-        unescaped.zip(escaped).map { |x,y|
+      escaped_url = Mechanize::Util.html_unescape(
+        url.split(/((?:%[0-9A-Fa-f]{2})+|#)/).each_slice(2).map { |x, y|
           "#{WEBrick::HTTPUtils.escape(x)}#{y}"
         }.join('')
       )
 
       begin
-        uri = URI.parse(escaped_uri)
+        uri = URI.parse(escaped_url)
       rescue
-        uri = URI.parse(WEBrick::HTTPUtils.escape(escaped_uri))
+        uri = URI.parse(WEBrick::HTTPUtils.escape(escaped_url))
       end
     end
 
     scheme = uri.relative? ? 'relative' : uri.scheme.downcase
     uri = @scheme_handlers[scheme].call(uri, referer)
 
-    if referer && referer.uri
+    if referer_uri
       if uri.path.length == 0 && uri.relative?
-        uri.path = referer.uri.path
+        uri.path = referer_uri.path
       end
     end
 
@@ -544,17 +588,16 @@ class Mechanize::HTTP::Agent
 
     if uri.relative?
       raise ArgumentError, "absolute URL needed (not #{uri})" unless
-        referer && referer.uri
+        referer_uri
 
-      base = nil
-      if referer.respond_to?(:bases) && referer.parser
-        base = referer.bases.last
+      if referer.respond_to?(:bases) && referer.parser &&
+          (lbase = referer.bases.last) && lbase.uri && lbase.uri.absolute?
+        base = lbase
+      else
+        base = nil
       end
 
-      uri = ((base && base.uri && base.uri.absolute?) ?
-             base.uri :
-             referer.uri) + uri
-      uri = referer.uri + uri
+      uri = referer_uri + (base ? base.uri : referer_uri) + uri
       # Strip initial "/.." bits from the path
       uri.path.sub!(/^(\/\.\.)+(?=\/)/, '')
     end
@@ -602,7 +645,11 @@ class Mechanize::HTTP::Agent
                             referer)
     raise Mechanize::UnauthorizedError, page unless @user || @password
 
-    challenges = @authenticate_parser.parse response['www-authenticate']
+    www_authenticate = response['www-authenticate']
+
+    raise Mechanize::UnauthorizedError, page unless www_authenticate
+
+    challenges = @authenticate_parser.parse www_authenticate
 
     if challenge = challenges.find { |c| c.scheme =~ /^Digest$/i } then
       realm = challenge.realm uri
@@ -631,7 +678,7 @@ class Mechanize::HTTP::Agent
       if challenge.params then
         type_2 = Net::NTLM::Message.decode64 challenge.params
 
-        type_3 = type_2.response({ :user => @user, :password => @password, },
+        type_3 = type_2.response({ :user => @user, :password => @password, :domain => @domain },
                                  { :ntlmv2 => true }).encode64
 
         headers['Authorization'] = "NTLM #{type_3}"
@@ -656,71 +703,45 @@ class Mechanize::HTTP::Agent
   end
 
   def response_content_encoding response, body_io
-    length = response.content_length
+    length = response.content_length ||
+      case body_io
+      when Tempfile, IO then
+        body_io.stat.size
+      else
+        body_io.length
+      end
 
-    length = case body_io
-             when IO, Tempfile then
-               body_io.stat.size
+    return body_io if length.zero?
+
+    out_io = case response['Content-Encoding']
+             when nil, 'none', '7bit' then
+               body_io
+             when 'deflate' then
+               content_encoding_inflate body_io
+             when 'gzip', 'x-gzip' then
+               content_encoding_gunzip body_io
              else
-               body_io.length
-             end unless length
-
-    out_io = nil
-
-    case response['Content-Encoding']
-    when nil, 'none', '7bit' then
-      out_io = body_io
-    when 'deflate' then
-      log.debug('deflate body') if log
-
-      return if length.zero?
-
-      begin
-        out_io = inflate body_io
-      rescue Zlib::BufError, Zlib::DataError
-        log.error('Unable to inflate page, retrying with raw deflate') if log
-        body_io.rewind
-        begin
-          out_io = inflate body_io, -Zlib::MAX_WBITS
-        rescue Zlib::BufError, Zlib::DataError
-          log.error("unable to inflate page: #{$!}") if log
-          nil
-        end
-      end
-    when 'gzip', 'x-gzip' then
-      log.debug('gzip body') if log
-
-      return if length.zero?
-
-      begin
-        zio = Zlib::GzipReader.new body_io
-        out_io = Tempfile.new 'mechanize-decode'
-        out_io.binmode
-
-        until zio.eof? do
-          out_io.write zio.read 16384
-        end
-      rescue Zlib::BufError, Zlib::GzipFile::Error
-        log.error('Unable to gunzip body, trying raw inflate') if log
-        body_io.rewind
-        body_io.read 10
-
-        out_io = inflate body_io, -Zlib::MAX_WBITS
-      rescue Zlib::DataError
-        log.error("unable to gunzip page: #{$!}") if log
-        ''
-      ensure
-        zio.close if zio and not zio.closed?
-      end
-    else
-      raise Mechanize::Error,
-            "Unsupported Content-Encoding: #{response['Content-Encoding']}"
-    end
+               raise Mechanize::Error,
+                 "unsupported content-encoding: #{response['Content-Encoding']}"
+             end
 
     out_io.flush
     out_io.rewind
 
     out_io
+  rescue Zlib::Error => e
+    message = "error handling content-encoding #{response['Content-Encoding']}:"
+    message << " #{e.message} (#{e.class})"
+    raise Mechanize::Error, message
+  ensure
+    begin
+      if Tempfile === body_io and
+         (StringIO === out_io or out_io.path != body_io.path) then
+        body_io.close! 
+      end
+    rescue IOError
+      # HACK ruby 1.8 raises IOError when closing the stream
+    end
   end
 
   def response_cookies response, uri, page
@@ -752,7 +773,8 @@ class Mechanize::HTTP::Agent
 
   def response_follow_meta_refresh response, uri, page, redirects
     delay, new_url = get_meta_refresh(response, uri, page)
-    return nil unless new_url
+    return nil unless delay
+    new_url = new_url ? resolve(new_url, page) : uri
 
     raise Mechanize::RedirectLimitReachedError.new(page, redirects) if
       redirects + 1 > @redirection_limit
@@ -760,7 +782,7 @@ class Mechanize::HTTP::Agent
     sleep delay
     @history.push(page, page.uri)
     fetch new_url, :get, {}, [],
-          Mechanize::Page.new(nil, {'content-type'=>'text/html'}), redirects
+          Mechanize::Page.new, redirects
   end
 
   def response_log response
@@ -778,11 +800,12 @@ class Mechanize::HTTP::Agent
     @context.parse uri, response, body_io
   end
 
-  def response_read response, request
+  def response_read response, request, uri
     content_length = response.content_length
 
-    if content_length and content_length > @max_file_buffer then
+    if use_tempfile? content_length then
       body_io = Tempfile.new 'mechanize-raw'
+      body_io.unlink
       body_io.binmode if defined? body_io.binmode
     else
       body_io = StringIO.new
@@ -795,9 +818,10 @@ class Mechanize::HTTP::Agent
       response.read_body { |part|
         total += part.length
 
-        if StringIO === body_io and total > @max_file_buffer then
+        if StringIO === body_io and use_tempfile? total then
           new_io = Tempfile.new 'mechanize-raw'
-          new_io.binmode if defined? binmode
+          new_io.unlink
+          new_io.binmode
 
           new_io.write body_io.string
 
@@ -809,7 +833,8 @@ class Mechanize::HTTP::Agent
       }
     rescue Net::HTTP::Persistent::Error => e
       body_io.rewind
-      raise Mechanize::ResponseReadError.new(e, response, body_io)
+      raise Mechanize::ResponseReadError.new(e, response, body_io, uri,
+                                             @context)
     end
 
     body_io.flush
@@ -846,9 +871,8 @@ class Mechanize::HTTP::Agent
 
     redirect_method = method == :head ? :head : :get
 
-    from_uri = page.uri
-    @history.push(page, from_uri)
-    new_uri = from_uri + response['Location'].to_s
+    @history.push(page, page.uri)
+    new_uri = resolve response['Location'].to_s, page
 
     fetch new_uri, redirect_method, {}, [], referer, redirects + 1
   end
@@ -907,78 +931,171 @@ class Mechanize::HTTP::Agent
 
   # :section: SSL
 
+  # Path to an OpenSSL CA certificate file
+  def ca_file
+    @http.ca_file
+  end
+
+  # Sets the path to an OpenSSL CA certificate file
+  def ca_file= ca_file
+    @http.ca_file = ca_file
+  end
+
+  # The SSL certificate store used for validating connections
+  def cert_store
+    @http.cert_store
+  end
+
+  # Sets the SSL certificate store used for validating connections
+  def cert_store= cert_store
+    @http.cert_store = cert_store
+  end
+
+  # The client X509 certificate
   def certificate
     @http.certificate
   end
 
+  # Sets the client certificate to given X509 certificate.  If a path is given
+  # the certificate will be loaded and set.
+  def certificate= certificate
+    certificate = if OpenSSL::X509::Certificate === certificate then
+                    certificate
+                  else
+                    OpenSSL::X509::Certificate.new File.read certificate
+                  end
+
+    @http.certificate = certificate
+  end
+
+  # An OpenSSL private key or the path to a private key
+  def private_key
+    @http.private_key
+  end
+
+  # Sets the client's private key
+  def private_key= private_key
+    private_key = if OpenSSL::PKey::PKey === private_key then
+                    private_key
+                  else
+                    OpenSSL::PKey::RSA.new File.read(private_key), @pass
+                  end
+
+    @http.private_key = private_key
+  end
+
+  # SSL version to use
+  def ssl_version
+    @http.ssl_version
+  end if RUBY_VERSION > '1.9'
+
+  # Sets the SSL version to use
+  def ssl_version= ssl_version
+    @http.ssl_version = ssl_version
+  end if RUBY_VERSION > '1.9'
+
+  # A callback for additional certificate verification.  See
+  # OpenSSL::SSL::SSLContext#verify_callback
+  #
+  # The callback can be used for debugging or to ignore errors by always
+  # returning +true+.  Specifying nil uses the default method that was valid
+  # when the SSLContext was created
+  def verify_callback
+    @http.verify_callback
+  end
+
+  # Sets the certificate verify callback
+  def verify_callback= verify_callback
+    @http.verify_callback = verify_callback
+  end
+
+  # How to verify SSL connections.  Defaults to VERIFY_PEER
+  def verify_mode
+    @http.verify_mode
+  end
+
+  # Sets the mode for verifying SSL connections
+  def verify_mode= verify_mode
+    @http.verify_mode = verify_mode
+  end
+
   # :section: Timeouts
 
-  # Sets the conection idle timeout for persistent connections
+  # Reset connections that have not been used in this many seconds
+  def idle_timeout
+    @http.idle_timeout
+  end
+
+  # Sets the connection idle timeout for persistent connections
   def idle_timeout= timeout
-    @idle_timeout = timeout
-    @http.idle_timeout = timeout if @http
+    @http.idle_timeout = timeout
   end
 
   # :section: Utility
 
+  ##
+  # Creates a new output IO by reading +input_io+ in +read_size+ chunks.  If
+  # the output is over the max_file_buffer size a Tempfile with +name+ is
+  # created.
+  #
+  # If a block is provided, each chunk of +input_io+ is yielded for further
+  # processing.
+
+  def auto_io name, read_size, input_io
+    out_io = StringIO.new
+
+    out_io.set_encoding Encoding::BINARY if out_io.respond_to? :set_encoding
+
+    until input_io.eof? do
+      if StringIO === out_io and use_tempfile? out_io.size then
+        new_io = Tempfile.new name
+        new_io.unlink
+        new_io.binmode
+
+        new_io.write out_io.string
+        out_io = new_io
+      end
+
+      chunk = input_io.read read_size
+      chunk = yield chunk if block_given?
+
+      out_io.write chunk
+    end
+
+    out_io.rewind
+
+    out_io
+  end
+
   def inflate compressed, window_bits = nil
     inflate = Zlib::Inflate.new window_bits
-    out_io = Tempfile.new 'mechanize-decode'
 
-    until compressed.eof? do
-      out_io.write inflate.inflate compressed.read 1024
+    out_io = auto_io 'mechanize-inflate', 1024, compressed do |chunk|
+      inflate.inflate chunk
     end
 
     out_io.write inflate.finish
 
     out_io
+  ensure
+    inflate.close
   end
 
   def log
     @context.log
   end
 
-  def set_http
-    @http = Net::HTTP::Persistent.new 'mechanize', @proxy_uri
-
-    @http.keep_alive = @keep_alive_time
-    @http.idle_timeout = @idle_timeout if @idle_timeout
-    @http.retry_change_requests = @retry_change_requests
-
-    @http.ca_file         = @ca_file
-    @http.cert_store      = @cert_store if @cert_store
-    @http.verify_callback = @verify_callback
-    @http.verify_mode     = @verify_mode if @verify_mode
-
-    # update our cached value
-    @verify_mode = @http.verify_mode
-    @cert_store  = @http.cert_store
-
-    if @cert and @key then
-      cert = if OpenSSL::X509::Certificate === @cert then
-               @cert
-             else
-               OpenSSL::X509::Certificate.new ::File.read @cert
-             end
-
-      key = if OpenSSL::PKey::PKey === @key then
-              @key
-            else
-              OpenSSL::PKey::RSA.new ::File.read(@key), @pass
-            end
-
-      @http.certificate = cert
-      @http.private_key = key
-    end
-  end
-
   ##
   # Sets the proxy address, port, user, and password +addr+ should be a host,
   # with no "http://", +port+ may be a port number, service name or port
   # number string.
 
-  def set_proxy(addr, port, user = nil, pass = nil)
-    return unless addr and port
+  def set_proxy addr, port, user = nil, pass = nil
+    unless addr and port then
+      @http.proxy = nil
+
+      return
+    end
 
     unless Integer === port then
       begin
@@ -992,12 +1109,19 @@ class Mechanize::HTTP::Agent
       end
     end
 
-    @proxy_uri = URI "http://#{addr}"
-    @proxy_uri.port = port
-    @proxy_uri.user     = user if user
-    @proxy_uri.password = pass if pass
+    proxy_uri = URI "http://#{addr}"
+    proxy_uri.port = port
+    proxy_uri.user     = user if user
+    proxy_uri.password = pass if pass
 
-    @proxy_uri
+    @http.proxy = proxy_uri
+  end
+
+  def use_tempfile? size
+    return false unless @max_file_buffer
+    return false unless size
+
+    size >= @max_file_buffer
   end
 
 end
