@@ -275,16 +275,45 @@ but not <a href="/" rel="me nofollow">this</a>!
     assert_equal 'GET', page.header['X-Request-Method']
   end
 
-  #def test_download
-  #  Dir.mktmpdir do |dir|
-  #    file = "#{dir}/download"
-  #    open file, 'w' do |io|
-  #      @mech.download 'http://example', io
-  #    end
+  def test_download
+    page = nil
 
-  #    assert_equal 1, File.stat(file).size
-  #  end
-  #end
+    in_tmpdir do
+      open 'download', 'w' do |io|
+        page = @mech.download 'http://example', io
+
+        refute io.closed?
+      end
+
+      assert_operator 1, :<=, File.stat('download').size
+    end
+
+    assert_empty @mech.history
+    assert_kind_of Mechanize::Page, page
+  end
+
+  def test_download_filename
+    page = nil
+
+    in_tmpdir do
+      page = @mech.download 'http://example', 'download'
+
+      assert_operator 1, :<=, File.stat('download').size
+    end
+
+    assert_empty @mech.history
+    assert_kind_of Mechanize::Page, page
+  end
+
+  def test_download_filename_error
+    in_tmpdir do
+      assert_raises Mechanize::UnauthorizedError do
+        @mech.download 'http://example/digest_auth', 'download'
+      end
+
+      refute File.exist? 'download'
+    end
+  end
 
   def test_get
     uri = URI 'http://localhost'
@@ -359,13 +388,6 @@ but not <a href="/" rel="me nofollow">this</a>!
     assert block_called
   end
 
-  def test_get_file
-    page = @mech.get("http://localhost/frame_test.html")
-    content_length = page.header['Content-Length']
-    page_as_string = @mech.get_file("http://localhost/frame_test.html")
-    assert_equal(content_length.to_i, page_as_string.length.to_i)
-  end
-
   def test_get_follow_meta_refresh
     @mech.follow_meta_refresh = true
 
@@ -667,6 +689,21 @@ but not <a href="/" rel="me nofollow">this</a>!
     assert_equal('File Upload Form', pages.title)
   end
 
+  def test_get_file
+    body = @mech.get_file 'http://localhost/frame_test.html'
+
+    assert_kind_of String, body
+    refute_empty body
+  end
+
+  def test_get_file_download
+    # non-Mechanize::File
+    body = @mech.get_file 'http://localhost/button.jpg'
+
+    assert_kind_of String, body
+    refute_empty body
+  end
+
   def test_head
     page = @mech.head('http://localhost/verb', { 'q' => 'foo' })
     assert_equal 0, @mech.history.length
@@ -728,6 +765,12 @@ but not <a href="/" rel="me nofollow">this</a>!
                  @mech.history[1].uri.to_s)
   end
 
+  def test_initialize
+    mech = Mechanize.new
+
+    assert_equal 50, mech.max_history
+  end
+
   def test_html_parser_equals
     @mech.html_parser = {}
     assert_raises(NoMethodError) {
@@ -735,10 +778,14 @@ but not <a href="/" rel="me nofollow">this</a>!
     }
   end
 
-  def test_idle_timeout_equals
-    @mech.idle_timeout = 5
+  def test_idle_timeout_default
+    assert_equal 5, Mechanize.new.idle_timeout
+  end
 
-    assert_equal 5, @mech.idle_timeout
+  def test_idle_timeout_equals
+    @mech.idle_timeout = 15
+
+    assert_equal 15, @mech.idle_timeout
   end
 
   def test_keep_alive_equals
@@ -893,6 +940,42 @@ but not <a href="/" rel="me nofollow">this</a>!
     assert_equal 5, @mech.read_timeout
   end
 
+  def test_referer
+    host_path = "localhost/tc_referer.html?t=1"
+    ['http', 'https'].each { |proto|
+      referer = "#{proto}://#{host_path}"
+      [
+        "",
+	"@",
+	"user1@",
+	":@",
+	"user1:@",
+	":password1@",
+	"user1:password1@",
+      ].each { |userinfo|
+        url = "#{proto}://#{userinfo}#{host_path}"
+        [url, url + "#foo"].each { |furl|
+          [
+            ['relative',		true],
+            ['insecure',		proto == 'http'],
+            ['secure',			true],
+            ['relative noreferrer',	false],
+            ['insecure noreferrer',	false],
+            ['secure noreferrer',	false],
+          ].each_with_index { |(type, bool), i|
+            rpage = @mech.get(furl)
+            page = rpage.links[i].click
+            assert_equal bool ? referer : '', page.body, "%s link from %s" % [type, furl]
+          }
+
+          rpage = @mech.get(furl)
+          page = rpage.forms.first.submit
+          assert_equal referer, page.body, "post from %s" % furl
+        }
+      }
+    }
+  end
+
   def test_retry_change_requests_equals
     refute @mech.retry_change_requests
 
@@ -911,7 +994,7 @@ but not <a href="/" rel="me nofollow">this</a>!
     assert_equal 'user',      @mech.proxy_user
     assert_equal 'pass',      @mech.proxy_pass
 
-    refute_same http, @mech.agent.http
+    assert_equal URI('http://user:pass@localhost:8080'), http.proxy_uri
   end
 
   def test_submit_bad_form_method
